import logging
from collections import Counter
from collections.abc import Mapping
from abc import abstractmethod
from string import ascii_uppercase

from bitarray import bitarray
from bitarray.util import ba2int, int2ba, ba2hex, zeros

from .util import rotate, chunk, readtsv, libroot


log = logging.getLogger(__name__)


class InvalidPassword(ValueError):
    def __str__(self):
        return "Invalid password: " + super().__str__()


class Field:
    """ Attributes of a password data field """
    def __init__(self, fid, offset, width, _type='uint', mod=0,
                 group='misc', order=0, desc='', **kwargs):
        if isinstance(offset, str):
            offset = int(offset, 0)
        if isinstance(width, str):
            width = int(width, 0)
        if isinstance(mod, str):
            mod = int(mod, 0) if mod else 0
        if isinstance(order, str):
            order = int(order) if order else 0
        if 'type' in kwargs:
            _type = kwargs['type']

        self.fid = fid
        self.offset = offset
        self.width = width
        self.group = group
        self.order = order
        self.desc = desc
        self.type = _type
        self.mod = mod

    @property
    def slice(self):
        """ The slice of the parent bitarray from which to read this field. """
        start = self.offset
        stop = start + self.width
        return slice(start, stop)

    @classmethod
    def gamefields(cls, game):
        path = f'{libroot}/game/{game}.tsv'
        fields = [Field(**record) for record in readtsv(path)]
        ct_fids = Counter(f.fid for f in fields)

        # Automagically append digits to the fids of duplicated fields
        # FIXME: There should be some way to make this explicit in the file
        # format.
        for fid, count in ct_fids.items():
            if count == 1:
                continue
            for i, field in enumerate((f for f in fields if f.fid == fid), 1):
                field.fid += str(i)

        return {f.fid: f for f in fields}


class Password(Mapping):
    _games = {}
    # This must be provided by subclasses
    gid = None      # Game ID
    default = None  # Default password

    def __init_subclass__(cls):
        if not cls.gid:
            raise ValueError(f"{cls} didn't specify game name")
        cls._games[cls.gid] = cls

    @classmethod
    def make(cls, gid, password=None, infile=None):
        pw = cls._games[gid](password)
        if infile:
            pw.load(infile)
        return pw

    @classmethod
    def supported_games(cls):
        return list(cls._games)

    @abstractmethod
    def __init__(self, password=None):
        """ Construct a password

        Subclasses must implement this. Init must accept a single, optional
        argument, `password`. If the password is not provided, or is None, a
        default must be used. The default password should produce the initial
        game state, if possible.
        """
        raise NotImplementedError

    @abstractmethod
    def __str__(self):
        """ Stringify a password

        Subclasses must implement this. It should print the password as it
        would be entered into the game.
        """
        raise NotImplementedError

    def load(self, f):
        for line in f:
            # Skip comments and blank lines
            if line.startswith("#") or not line.strip():
                continue
            k, v = (part.strip() for part in line.split(":"))
            self[k] = int(v, 0)

    def dump(self):
        out = ''
        colw = max(len(k) for k in self) + 2
        for key, val in self.items():
            key = key + ":"
            val = int(val)
            out += f'{key:{colw}}{val}\n'
        return out



class Structure(Mapping):
    fields = {}
    _initialized = False

    def __init__(self):
        # Subclasses should run super().__init__ *after* doing their own
        # initialization
        self.data = zeros(self.ct_bits, 'little')
        self._initialized = True

    @property
    def ct_bits(self):
        return sum(f.width for f in self.fields.values())

    @property
    def hex(self):
        return ba2hex(self.data)

    def __iter__(self):
        return iter(self.fields)

    def __len__(self):
        return len(self.fields)

    def __getattr__(self, name):
        if name not in self.fields:
            raise AttributeError(f"No such field: {name}")
        field = self.fields[name]
        return ba2int(self.data[field.slice]) + field.mod

    def __setattr__(self, name, value):
        if not self._initialized:
            object.__setattr__(self, name, value)
        elif name in self.fields:
            field = self.fields[name]
            value = int(value) - field.mod
            length = field.width
            end = self.data.endian()
            self.data[field.slice] = int2ba(value, length=length, endian=end)
        else:
            super().__setattr__(name, value)

    def __getitem__(self, name):
        if name not in self.fields and not hasattr(self, name):
            raise KeyError(f"No such field: {name}")
        return getattr(self, name)

    def __setitem__(self, name, value):
        if name not in self.fields and not hasattr(self, name):
            raise KeyError(f"No such field: {name}")
        setattr(self, name, value)


class MetroidPassword(Structure, Password):
    gid = 'metroid'
    fields = Field.gamefields(gid)

    def __init__(self, password=None):
        if password is None:
            password = '0' * 24
        password = password.strip().replace(' ', '')  # Remove spaces
        if len(password) != 24:
            raise InvalidPassword("wrong length")

        data = bitarray(endian='big')
        for charcode in password.encode(self.gid):
            data += int2ba(charcode, 6, endian='big')

        self.shift = ba2int(data[-16:-8])
        gamestate = rotate(data[:-16], -self.shift)
        chk_given = ba2int(data[-8:])
        chk_data = (sum(gamestate.tobytes()) + self.shift) % 0x100
        if chk_given != chk_data:
            raise InvalidPassword(f"checksum failure, {chk_given} != {chk_data}")
        self.data = bitarray(gamestate, 'little')

        self._initialized = True

    @property
    def checksum(self):
        return (sum(self.data.tobytes()) + self.shift) % 0x100

    @property
    def codepoints(self):
        # The documentation's bit-numbering is lsb0, but slices for the code
        # points are based on msb0. Hence making a new bitarray with the
        # opposite endianness. I *think* the right thing to do here is to
        # re-index the field offsets according to msb0, but I'm not sure. Try
        # it and see what the spec looks like.
        bits = rotate(bitarray(self.data, 'big'), self.shift)
        bits += int2ba(self.shift, 8)
        bits += int2ba(self.checksum, 8)
        assert len(bits) == 144

        for c in chunk(bits, 6):
            yield ba2int(c)

    def __str__(self):
        pw = bytes(self.codepoints).decode(self.gid)
        return ' '.join(chunk(pw, 6))


class KidIcarusPassword(MetroidPassword):
    gid = 'ki'
    fields = Field.gamefields(gid)

    def __init__(self, password=None):
        if password is None:
            password = '0' * 24
        password = password.strip().replace(' ', '')  # Remove spaces
        if len(password) != 24:
            raise InvalidPassword("wrong length")

        data = bitarray(endian='little')
        for charcode in password.encode(self.gid):
            data += int2ba(charcode, 6, endian='little')
        self.data = bitarray(data[:-8], 'little')
        checksum = ba2int(data[-8:])

        if self.checksum != checksum:
            msg = f"checksum failure, {self.checksum} != {checksum} [{password}]"
            raise InvalidPassword(msg)
        self._initialized = True

    @property
    def level(self):
        sub = 4 if self.maze else self.substage
        return f'{self.stage}-{sub}'

    @level.setter
    def level(self, value):
        stage, substage = (int(part) for part in value.split('-'))
        self.stage = stage
        self.substage = 0 if substage == 4 else substage
        self.maze = int(substage == 4)

    @property
    def bits(self):
        return self.data + int2ba(self.checksum, 8, 'little')

    @property
    def checksum(self):
        return sum(self.data.tobytes()) % 0x100

    @property
    def codepoints(self):
        for c in chunk(bitarray(self.bits, 'little'), 6):
            yield ba2int(c)

class SolarJetmanPassword(Structure, Password):
    # this works like hex with a different alphabet
    charset = 'BDGHKLMNPQRTVWXZ'
    defaultpw = 'BBBBBBBBBBBB'
    gid = 'sj'
    fields = Field.gamefields(gid)

    def __init__(self, password=None):
        super().__init__()
        if password is None:
            password = self.defaultpw
        for i, char in enumerate(password):
            start = i * 4
            end = start + 4
            cp = self.charset.index(char)
            self.data[start:end] = int2ba(cp, 4, 'little')
        assert len(self.data) == self.ct_bits

    @property
    def score(self):
        score = 0
        for exp in range(6):
            digit = self[f'score{exp}']
            score += digit * 10 ** exp
        return score

    @score.setter
    def score(self, value):
        for exp in range(6):
            digit = (value // 10 ** exp) % 10
            self[f'score{exp}'] = digit

    @property
    def level(self):
        return (self.levelhigh << 2) + self.levellow + 1

    @level.setter
    def level(self, value):
        value -= 1
        self.levelhigh = value // 4
        self.levellow = value % 4

    def __iter__(self):
        yield 'level'
        yield 'score'
        for fid, field in self.fields.items():
            if field.order >= 0:
                yield fid

    def __len__(self):
        return len(list(self))

    @property
    def codepoints(self):
        cp = [ba2int(ck) for ck in chunk(self.data, 4)]

        # The checksum algorithm is made of horrible...I feel like there should
        # be a cleaner way to represent this.
        chk1 = ((cp[0] ^ cp[1]) + cp[2] ^ cp[4]) + cp[5]
        chk2 = ((cp[6] ^ cp[7]) + cp[8] ^ cp[10]) + cp[11]

        chk2 += int(chk1 >= 16)
        chk1 += chk2 // 16

        cp[3] = chk1 % 16
        cp[9] = chk2 % 16
        return cp

    def __str__(self):
        return ''.join(self.charset[cp] for cp in self.codepoints)



class MM2Boss:
    def __init__(self, name, alive, dead):
        self.name = name
        self.alive = int(alive)
        self.dead = int(dead)


class MM2Password(Password):
    gid = 'mm2'
    bosses = {record['name']: MM2Boss(**record)
              for record in readtsv(f'{libroot}/game/mm2.tsv')}

    def __init__(self, password=None):
        self.tanks = None
        self.defeated = dict()
        for boss in self.bosses:
            self.defeated[boss] = None

        if password:
            self._load_password(password)
        else:
            self._set_defaults()
        assert self.tanks is not None
        assert not any(state is None for state in self.defeated.values())
        self._initialized = True

    def _set_defaults(self):
        self.tanks = 0
        for name in self.bosses:
            self.defeated[name] = False

    def _load_password(self, password):
        codes = sorted(self.cell2int(cell) for cell in password.split(" "))
        if len(codes) != 9:
            raise InvalidPassword("Wrong length")
        if codes[0] > 4:
            raise InvalidPassword("Tank cell missing")

        self.tanks = codes[0]
        codes = [(code - 5 - self.tanks) % 20
                 for code in codes[1:]]
        for boss in self.bosses.values():
            alive = boss.alive in codes
            dead = boss.dead in codes
            if alive and dead:
                raise InvalidPassword(f"{boss.name} is schrodinger's boss")
            elif alive:
                self[boss.name] = False
            elif dead:
                self[boss.name] = True
            else:
                raise InvalidPassword(f"No state cell for {boss.name}")

    @staticmethod
    def cell2int(cell):
        if len(cell) != 2:
            raise ValueError("Cells are letter-number pairs")
        row = cell[0]
        col = cell[1]
        return (ord(row) - ord('A')) * 5 + (int(col)-1)

    @staticmethod
    def int2cell(i):
        return ascii_uppercase[i // 5] + str(i % 5 + 1)


    def __str__(self):
        codepoints = [self.tanks]
        for boss in self.bosses.values():
            base = boss.dead if self[boss.name] else boss.alive
            code = (base + self.tanks) % 20 + 5
            codepoints.append(code)
        return ' '.join(sorted(self.int2cell(cp) for cp in codepoints))

    def __len__(self):
        return len(self.defeated) + 1

    def __iter__(self):
        yield 'tanks'
        yield from self.bosses.keys()

    def __getitem__(self, k):
        if k == 'tanks':
            return self.tanks
        else:
            return self.defeated[k]

    def __setitem__(self, k, v):
        v = int(v)
        if k == 'tanks':
            self.tanks = v
        else:
            self.defeated[k] = v



from unittest import TestCase, expectedFailure
from functools import partial

from retropass import Password

class TestMetroid(TestCase):
    def setUp(self):
        self.make = partial(Password.make, 'metroid')
        self.default = self.make()

    def test_password_roundtrip(self):
        passwords = ['000000 000000 000000 000000',
                     '0G0000 000000 400000 00000H',
                     'GGW01G 000020 VsG000 00002n',
                     'ENGAGE RIDLEY MOTHER FUCKER',]
        for text in passwords:
            pw = self.make(text)
            self.assertEqual(text, str(pw))

    def test_known_password_1(self):
        pw = self.make()
        pw.taken_marumari = True
        pw.has_marumari = True
        self.assertEqual(str(pw), '0G0000 000000 400000 00000H')

    def test_known_password_2(self):
        pw = self.make()
        pw.has_bombs = 1
        pw.has_boots = 1
        pw.has_longbeam = 1
        pw.has_marumari = 1
        pw.has_screw = 1
        pw.has_varia = 1
        pw.has_wave = 1
        pw.missiles = 100
        pw.taken_bombs = 1
        pw.taken_boots = 1
        pw.taken_marumari = 1
        pw.taken_screw = 1
        pw.taken_varia = 1
        pw.unarmored = 1
        print(pw.checksum)
        self.assertEqual(str(pw), 'GGW01G 000020 VsG000 00002n')

    def test_known_password_3(self):
        pw = self.make('ENGAGE RIDLEY MOTHER FUCKER')
        self.assertEqual(pw.ridley_dead, 1)
        self.assertEqual(pw.missiles, 82)

    def test_data_length(self):
        self.assertEqual(len(self.default.data), 128)

    def test_pw_length(self):
        self.assertEqual(len(str(self.default)), 24+3)  # +3 for the spaces

class TestSolarJetman(TestCase):
    def setUp(self):
        self.make = partial(Password.make, 'sj')
        self.default = self.make()

    def test_password_roundtrip(self):
        passwords = ['ZQBQQQQQHGTQ',
                     'ZQBQQQQQNMTQ',
                     'ZQBPQQQQTRTQ']
        for text in passwords:
            pw = self.make(text)
            self.assertEqual(text, str(pw))

    def test_known_passwords(self):
        pw = self.make()
        self.assertEqual(str(pw), "BBBBBBBBBBBB")
        pw.level = 6
        self.assertEqual(str(pw), "BBDDBBBBKKBB")
        pw.score = 123456
        self.assertEqual(str(pw), "BMDPKLHGKMBD")
        pw.lives = 3
        self.assertEqual(str(pw), "HMDNKLHGKMBD")
        pw.shields = 1
        self.assertEqual(str(pw), "HMDNKLHGKPGD")
        pw.map = 1
        self.assertEqual(str(pw), "HMDNKLHGMMGD")
        pw.thrusters = 1
        self.assertEqual(str(pw), "HMDNKLHGMLHD")


class TestKidIcarus(TestCase):
    """ By level:
    000000 000000 E30000 0000Yx: 1-1
    000000 000000 m20000 00002q: 1-1
    0000eu 60j700 uG0004 1000J0: 1-4
    0000ys T0J300 m2001C H000aS: 2-1
    00008C i04400 mIG01D I0005F: 2-4
    0000mu w0K200 O3G00H I100s5: 3-1
    0000y0 11X200 u0G00H I100t0: 3-4
    00008p 414100 O3G00H I500eB: 4-1


    """
    def setUp(self):
        self.make = partial(Password.make, 'ki')
        self.default = self.make()

    def test_password_roundtrip(self):
        passwords = ['000000 000000 000000 000000',
                     '000000 000000 000000 0004G0',
                     '000000 000000 E30000 0000Yx']
        for text in passwords:
            pw = self.make(text)
            self.assertEqual(text, str(pw))

    def test_known_passwords(self):
        text = '000000 000000 E30000 0000Yx'
        pw = self.make(text)
        self.assertEqual(text, str(pw))

    def test_known_password_levels(self):
        passwords = {
            "000000 000000 E30000 0000Yx": "1-1",
            "000000 000000 m20000 00002q": "1-1",
            "000000 000000 O10000 0004IU": "1-2",
            "0000aC 10v000 O50000 0004Yw": "1-2",
            "6eW3!! !!!!00 F38W!H C0042N": "1-2",
            "0000eu 60j700 uG0004 1000J0": "1-4",
            "0000ys T0J300 m2001C H000aS": "2-1",
            "C0G00Q G0v500 kO0008 I004af": "2-2",
            "C0W0C? K0e300 ET0008 I0084Z": "2-3",
            "00008C i04400 mIG01D I0005F": "2-4",
            "0000mu w0K200 O3G00H I100s5": "3-1",
            "CeW0qB d0j400 O3000C J104s7": "3-2",
            "EeW1yw h08300 u2000C J1086j": "3-3",
            "0000y0 11X200 u0G00H I100t0": "3-4",
            "00008p 414100 O3G00H I500eB": "4-1",
            }
        from bitarray.util import int2ba
        for text, level in passwords.items():
            pw = self.make(text)
            self.assertEqual(pw.level, level)
            self.assertEqual(str(pw), text)


class TestMM2(TestCase):
    def setUp(self):
        self.make = partial(Password.make, 'mm2')
        self.default = self.make()

    def test_known_password_1(self):
        # Starting state. No tanks, all bosses alive.
        text = "A1 B5 C3 C4 D2 D5 E1 E2 E4"
        pw = self.make(text)
        self.assertEqual(pw.tanks, 0)
        for boss in pw.bosses:
            self.assertFalse(pw[boss])
        self.assertEqual(text, str(pw))

    def test_known_password_2(self):
        # Four tanks and all bosses dead
        text = "A5 B2 B4 C1 C3 C5 D4 D5 E2"
        pw = self.make(text)
        self.assertEqual(pw.tanks, 4)
        for boss in pw.bosses:
            self.assertTrue(pw[boss])
        self.assertEqual(text, str(pw))



import os
import sys
import csv
import enum
from enum import Enum
from os.path import dirname, realpath


libroot = dirname(realpath(__file__))
csv.register_dialect(
        'tsv',
        delimiter='\t',
        lineterminator=os.linesep,
        quoting=csv.QUOTE_NONE,
        strict=True,
        )


def readtsv(path):
    with open(path, newline='') as f:
        return list(csv.DictReader(f, dialect='tsv'))


class bitorder(Enum):
    lsb0 = enum.auto()
    msb0 = enum.auto()


class byteorder(Enum):
    """ An enum for byte order, because I hate inline strings

    str()ing this will get you just the string. This is so it can be used
    directly as the byteorder arg to int.from_bytes, etc.
    """

    little = 'little'
    big = 'big'
    native = sys.byteorder

    def __str__(self):
        return self.value


# Convenience aliases
lsb0 = bitorder.lsb0
msb0 = bitorder.msb0
be = byteorder.big
le = byteorder.little
ne = byteorder.native

class BitList:
    def __init__(self, parent, offsets):
        self.parent = parent
        self.offsets = list(offsets)

    def __getitem__(self, i):
        return self.parent[self.offsets[i]]

    def __setitem__(self, i, v):
        self.parent[self.offsets[i]] = v

def rotate(bits, n):
    """ Rotate a bitarray n bits

    Positive values of N rotate right; negative values rotate left.
    """
    return bits[-n:] + bits[:-n]

def chunk(bits, chunk_size):
    for i in range(0, len(bits), chunk_size):
        yield bits[i:i+chunk_size]



from unittest import TestCase
from string import digits, ascii_uppercase, ascii_lowercase

import retropass.text as text

class TestCodecs(TestCase):
    def test_basics(self):
        codepoints = bytes(range(64))
        alpha = digits + ascii_uppercase + ascii_lowercase + '?-'
        self.assertEqual(alpha.encode('metroid'), codepoints)
        self.assertEqual(codepoints.decode('metroid'), alpha)

    def test_roundtrip(self):
        s = 'thisisastring'
        enc = 'metroid'
        self.assertEqual(s, s.encode(enc).decode(enc))



import sys
import argparse
import logging
from functools import partial
from argparse import FileType

import retropass as rp


log = logging.getLogger(__name__)


def main(argv=None):
    if argv is None:
        argv = sys.argv[1:]

    desc = "retro game password generator"

    # Check for the version option before normal parsing, so the parser doesn't
    # choke on missing args:
    if "-V" in argv or "--version" in argv:
        print('retropass v' + rp.version)
        sys.exit()

    parser = argparse.ArgumentParser(description=desc)

    addarg = parser.add_argument
    addopt = partial(addarg, nargs='?')
    addflag = partial(addarg, action='store_true')

    addarg("game", choices=rp.Password.supported_games(),
           help="game to generate password for")
    addopt("-c", "--conf", help="file to take settings from", type=FileType())
    addopt("-p", "--password", help="password to decode")
    addflag("-d", "--dump", help="dump resulting settings, or defaults")
    addflag("-v", "--verbose", help="verbose logging")
    addflag("-D", "--debug", help="debug logging")
    addflag("-V", "--version", help="print program version")

    args = parser.parse_args(argv)
    level = (logging.DEBUG if args.debug
             else logging.INFO if args.verbose
             else logging.WARN)
    lfmt = '%(levelname)s\t%(module)s:%(lineno)d\t%(message)s'
    logging.basicConfig(level=level, format=lfmt)
    log.debug("debug logging on")
    log.info("verbose logging on")

    try:
        pw = rp.Password.make(args.game, args.password, args.conf)
    except rp.InvalidPassword as ex:
        log.error(ex)
        sys.exit(2)

    if args.dump or args.password:
        print(pw.dump())
    else:
        print(pw)

if __name__ == '__main__':
    main()



from unittest import TestCase
from functools import partial

from bitarray import bitarray

import retropass.util as util
from retropass.util import rotate, chunk

msb = partial(bitarray, endian='big')
lsb = partial(bitarray, endian='little')

class TestRotate(TestCase):
    def test_noop_rotation(self):
        bits = msb('1000')
        self.assertEqual(rotate(bits, 0), bits)

    def test_rotate_right_big(self):
        bits = msb('1000')
        self.assertEqual(rotate(bits, 1), msb('0100'))

    def test_rotate_left_big(self):
        bits = msb('1000')
        self.assertEqual(rotate(bits, -1), msb('0001'))

    def test_rotate_right_little(self):
        bits = lsb('1000')
        self.assertEqual(rotate(bits, 1), lsb('0100'))

    def test_rotate_left_little(self):
        bits = lsb('1000')
        self.assertEqual(rotate(bits, -1), lsb('0001'))

    def test_rotate_wrap(self):
        bits = msb('1000')
        self.assertEqual(rotate(bits, 4), bits)

    def test_rotate_cancellation(self):
        for ba in lsb, msb:
            bits = ba('100100100100100100')
            rotated = rotate(bits, -1)
            rotated = rotate(rotated, 1)
            self.assertEqual(bits, rotated)

class TestChunk(TestCase):
    def test_chunk(self):
        bits = bitarray('100100100100')
        chunks = list(util.chunk(bits, 3))
        self.assertEqual(len(chunks), 4)
        for chunk in chunks:
            self.assertIsInstance(chunk, bitarray)
            self.assertEqual(len(chunk), 3)
    def test_chunk_length(self):
        bits = bitarray('10000000' * 18)
        chunks = list(util.chunk(bits, 6))
        self.assertEqual(len(chunks), 24)



from os.path import join, dirname

from setuptools import setup, find_packages

deps = ['bitstring>=3.1.3',
        'bitarray>=1.5.0',
        'pyyaml>=3.10']
scm_version_options = {
        'write_to': 'src/retropass/version.py',
        'fallback_version': 'UNKNOWN',
        }
classifiers = ['Development Status :: 4 - Beta',
               'Intended Audience :: Developers',
               'Natural Language :: English',
               'Operating System :: OS Independent',
               'Programming Language :: Python :: 3',
               'Topic :: Software Development :: Libraries :: Python Modules']


def read(filename):
    with open(join(dirname(__file__), filename)) as f:
        return f.read()

setup(name='retropass',
      packages=find_packages('src'),
      package_dir={'': 'src'},
      include_package_data=True,
      install_requires=deps,
      # FIXME: Pretty sure this installs scm outside pip. Maybe better to
      # encase setup in a try block and print an error suggesting it be
      # installed.
      setup_requires=['setuptools-scm>=3.3.0'],
      use_scm_version=scm_version_options,
      tests_require=['tox'],
      author='Andrew Vant',
      author_email='ajvant@gmail.com',
      classifiers=classifiers,
      description='Library and tool for generating progress passwords for old games',
      long_description=read('README.md'),
      long_description_content_type='text/markdown',
      zip_safe=False,
      keywords="rom roms nes",
      url="https://pypi.org/project/retropass/",
      test_suite='tests',
      entry_points={"console_scripts": ["retropass = retropass.cli:main"]},
      )



import codecs
import logging
from functools import lru_cache

from .util import libroot

log = logging.getLogger(__name__)

def load(name, f):
    """ Load a text codec from a file-like object"""

    c2b={}
    b2c={}

    for line in f:
        if not line.strip() or line.startswith('#'):
            continue
        line = line.rstrip('\n')
        byte, char = line.split('=')
        byte = int(byte, 16)
        c2b[char] = byte
        b2c[byte] = char

    assert len(c2b) == len(b2c)
    if not c2b:
        raise ValueError(f"text table for {name} appears to be empty")

    def decode(_bytes):
        s = ''.join(b2c[byte] for byte in _bytes)
        return (s, len(_bytes))

    def encode(s):
        _bytes = bytes(c2b[char] for char in s)
        return (_bytes, len(s))

    ci = codecs.CodecInfo(encode, decode)
    return ci


@lru_cache
def lookup(name):
    """ Look up a text codec by name

    This is registered as a codec search function, so it shouldn't need to be
    called directly."""

    path = f'{libroot}/game/{name}.tbl'
    try:
        with open(path) as f:
            return load(name, f)
    except FileNotFoundError as ex:
        log.debug(ex)
    return None

codecs.register(lookup)
